from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from . import urls


class DensMenuItem(MenuItemHook):
    """This class ensures only authorized users will see the menu entry"""

    def __init__(self):
        # setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            _("Dens"),
            "fas fa-tents fa-fw",
            "dens:index",
            navactive=["dens:"],
        )

    def render(self, request):
        if request.user.has_perm("dens.basic_access"):
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    return DensMenuItem()


@hooks.register("url_hook")
def register_urls():
    return UrlHook(urls, "dens", r"^dens/")


@hooks.register("charlink")
def register_charlink_hook():
    return "dens.thirdparty.charlink_hook"



import datetime
from copy import deepcopy

from structuretimers.models import Timer as StructureTimer

from django.test import TestCase
from django.utils import timezone
from eveuniverse.models import EveType

from allianceauth.eveonline.models import EveCharacter
from allianceauth.timerboard.models import Timer as TimerboardTimer

from dens.models import MercenaryDenReinforcedNotification
from dens.tests.utils import (
    NO_ALLIANCE_REINFORCE_NOTIFICATION,
    REINFORCE_NOTIFICATION,
    create_fake_den,
    create_fake_den_owner,
    create_fake_den_owner_no_alliance,
    create_fake_notification,
)

from ..tasks import DenMissingFromDatabase, create_reinforce_notification
from ..thirdparty.structuretimers import create_structuretimer_timer
from ..thirdparty.timerboard import create_timerboard_timer
from .testdata.load_eveuniverse import load_eveuniverse


class TestNotification(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_parsing(self):

        res = MercenaryDenReinforcedNotification.objects.parse_information_from_notification(
            REINFORCE_NOTIFICATION
        )

        self.assertIsNotNone(res)

        self.assertEqual(res["aggressor_alliance"]["id"], "99003214")
        self.assertEqual(res["aggressor_alliance"]["name"], "Brave Collective")
        self.assertEqual(res["aggressor_corporation"]["id"], "98169165")
        self.assertEqual(res["aggressor_corporation"]["name"], "Brave Newbies Inc.")
        self.assertEqual(res["aggressor_character_id"], 96914524)
        self.assertEqual(res["planet_id"], 40255101)
        self.assertEqual(res["solar_system_id"], 30004028)
        self.assertEqual(res["timestamp_entered"], 133761583913385305)
        self.assertEqual(res["timestamp_exited"], 133762405913385305)

    def test_create_from_notification(self):

        owner = create_fake_den_owner()
        create_fake_den(owner)

        create_reinforce_notification(REINFORCE_NOTIFICATION)

        stored_reinforce_notification = (
            MercenaryDenReinforcedNotification.objects.all()[0]
        )

        self.assertIsNotNone(stored_reinforce_notification)

        self.assertEqual(stored_reinforce_notification.id, 2064665856)
        self.assertEqual(
            stored_reinforce_notification.reinforced_by,
            EveCharacter.objects.get_character_by_id(96914524),
        )

        self.assertTrue(
            timezone.is_aware(stored_reinforce_notification.enter_reinforcement)
        )
        self.assertTrue(
            timezone.is_aware(stored_reinforce_notification.exit_reinforcement)
        )

    def test_in_future(self):
        reinforce_notification = create_fake_notification()

        self.assertFalse(reinforce_notification.is_in_future())

        reinforce_notification.exit_reinforcement = timezone.now() + datetime.timedelta(
            days=1
        )
        reinforce_notification.save()

        self.assertTrue(reinforce_notification.is_in_future())

    def test_add_timerboard(self):
        reinforce_notification = create_fake_notification()

        self.assertEqual(TimerboardTimer.objects.count(), 0)

        create_timerboard_timer(reinforce_notification)

        self.assertEqual(TimerboardTimer.objects.count(), 1)

        timer = TimerboardTimer.objects.all()[0]

        self.assertEqual(timer.details, "Mercenary den reinforced by Butt Chili")
        self.assertEqual(timer.system, "E-VKJV")
        self.assertEqual(timer.planet_moon, "E-VKJV VI")
        self.assertEqual(timer.structure, TimerboardTimer.Structure.MERCDEN)
        self.assertEqual(timer.timer_type, TimerboardTimer.TimerType.FINAL)
        self.assertEqual(timer.objective, TimerboardTimer.Objective.FRIENDLY)
        self.assertEqual(timer.eve_time, reinforce_notification.exit_reinforcement)
        self.assertFalse(timer.important)
        self.assertIsNone(timer.eve_character)
        self.assertEqual(
            timer.eve_corp,
            reinforce_notification.den.owner.character_ownership.character.corporation,
        )
        self.assertFalse(timer.corp_timer)
        self.assertIsNone(timer.user)

    def test_add_structuretimer(self):
        reinforce_notification = create_fake_notification()

        self.assertEqual(StructureTimer.objects.count(), 0)

        create_structuretimer_timer(reinforce_notification)

        self.assertEqual(StructureTimer.objects.count(), 1)

        timer = StructureTimer.objects.all()[0]

        structure_type, _ = EveType.objects.get_or_create_esi(id=85230)

        self.assertEqual(timer.date, reinforce_notification.exit_reinforcement)
        self.assertIsNone(timer.details_image_url)
        self.assertEqual(timer.details_notes, "Reinforced by Butt Chili")
        # self.assertEqual(timer.eve_alliance, 99011247)
        self.assertEqual(
            timer.eve_character,
            reinforce_notification.den.owner.character_ownership.character,
        )
        self.assertEqual(
            timer.eve_corporation,
            reinforce_notification.den.owner.character_ownership.character.corporation,
        )
        self.assertEqual(
            timer.eve_solar_system, reinforce_notification.den.location.eve_solar_system
        )
        self.assertFalse(timer.is_important)
        self.assertFalse(timer.is_opsec)
        self.assertEqual(timer.location_details, "E-VKJV VI")
        self.assertEqual(timer.objective, StructureTimer.Objective.FRIENDLY)
        self.assertEqual(timer.owner_name, "Ji'had Rokym")
        self.assertEqual(timer.structure_type, structure_type)
        self.assertEqual(timer.structure_name, "VI")
        self.assertEqual(timer.timer_type, StructureTimer.Type.FINAL)
        self.assertIsNone(timer.user)
        self.assertEqual(timer.visibility, StructureTimer.Visibility.UNRESTRICTED)

    def test_add_structuretimer_without_alliance(self):
        owner = create_fake_den_owner_no_alliance()
        create_fake_den(owner)
        create_reinforce_notification(REINFORCE_NOTIFICATION)
        notification = MercenaryDenReinforcedNotification.objects.all()[0]

        create_structuretimer_timer(notification=notification)

    def test_bad_notification(self):
        """Test when a notification that doesn't have the correct format is passed"""
        reinforce_notification = {"text": "BAD INPUT"}

        self.assertRaises(
            ValueError,
            MercenaryDenReinforcedNotification.objects.create_from_notification,
            reinforce_notification,
        )

    def test_past_notification_not_existing_den(self):
        """Receives a past notification for a den no longer existing"""
        create_reinforce_notification(REINFORCE_NOTIFICATION)

        self.assertEqual(MercenaryDenReinforcedNotification.objects.count(), 1)

        stored_reinforce_notification = (
            MercenaryDenReinforcedNotification.objects.all()[0]
        )

        self.assertEqual(stored_reinforce_notification.id, 2064665856)
        self.assertEqual(
            stored_reinforce_notification.reinforced_by,
            EveCharacter.objects.get_character_by_id(96914524),
        )

        self.assertTrue(
            timezone.is_aware(stored_reinforce_notification.enter_reinforcement)
        )
        self.assertTrue(
            timezone.is_aware(stored_reinforce_notification.exit_reinforcement)
        )

        self.assertIsNone(stored_reinforce_notification.den)

    def test_future_notification_not_existing_den(self):
        """
        Receives a notification for a den not yet out of reinforcement but not yet in the database
        Should occur if a den is reffed after anchoring before an asset update triggered
        """
        json_notification = deepcopy(
            REINFORCE_NOTIFICATION
        )  # Seems like not copying will override the value in all tests
        # Changes exit time to 2056
        json_notification["text"] = (
            'aggressorAllianceName: <a href="showinfo:16159//99003214">Brave Collective</a>\naggressorCharacterID: 96914524\naggressorCorporationName: <a href="showinfo:2//98169165">Brave Newbies Inc.</a>\nitemID: &id001 1047210542765\nmercenaryDenShowInfoData:\n- showinfo\n- 85230\n- *id001\nplanetID: 40255101\nplanetShowInfoData:\n- showinfo\n- 11\n- 40255101\nsolarsystemID: 30004028\ntimestampEntered: 133761583913385305\ntimestampExited: 143762405913385305\ntypeID: 85230\n'
        )

        self.assertRaises(
            DenMissingFromDatabase, create_reinforce_notification, json_notification
        )

    def test_correctly_parses_when_aggressor_not_in_alliance(self):
        self.assertIsNotNone(
            MercenaryDenReinforcedNotification.objects.parse_information_from_notification(
                NO_ALLIANCE_REINFORCE_NOTIFICATION
            )
        )

    def test_notification_id002(self):
        """
        This notification failed to parse because some additionnal fields were present in the text.
        `&id002` and `*id002`
        """
        notification = {
            "is_read": True,
            "notification_id": 2101785272,
            "sender_id": 1000438,
            "sender_type": "corporation",
            "text": 'aggressorAllianceName: <a href="showinfo:16159//99012770">Black Rose.</a>\naggressorCharacterID: 1056152916\naggressorCorporationName: <a href="showinfo:2//98772518">Orbital Bloom</a>\nitemID: &id001 1047787968947\nmercenaryDenShowInfoData:\n- showinfo\n- 85230\n- *id001\nplanetID: &id002 40128505\nplanetShowInfoData:\n- showinfo\n- 11\n- *id002\nsolarsystemID: 30002015\ntimestampEntered: 133814418228106194\ntimestampExited: 133815271538106194\ntypeID: 85230\n',
            "timestamp": datetime.datetime(
                2025, 1, 15, 19, 10, tzinfo=datetime.timezone.utc
            ),
            "type": "MercenaryDenReinforced",
        }

        MercenaryDenReinforcedNotification.objects.create_from_notification(
            notification
        )

    def test_backslash_n_in_corp_name_notification(self):
        """
        Notifications returned from the ESI can have a \n in the text field that breaks the previous regex parsing.
        Switching to a yaml parser should have fixed the problem
        """

        notification = {
            "notification_id": 2094252961,
            "sender_id": 1000438,
            "sender_type": "corporation",
            "text": 'aggressorAllianceName: <a href="showinfo:16159//1900696668">The Initiative.</a>\naggressorCharacterID: 96104294\naggressorCorporationName: <a href="showinfo:2//98457033">United Mining and Hauling\n  Inc</a>\nitemID: &id001 1047702713146\nmercenaryDenShowInfoData:\n- showinfo\n- 85230\n- *id001\nplanetID: 40255900\nplanetShowInfoData:\n- showinfo\n- 11\n- 40255900\nsolarsystemID: 30004041\ntimestampEntered: 133803021282336070\ntimestampExited: 133803877282336070\ntypeID: 85230\n',
            "timestamp": datetime.datetime(
                2025, 1, 2, 14, 35, tzinfo=datetime.timezone.utc
            ),
            "type": "MercenaryDenReinforced",
            "is_read": None,
        }

        MercenaryDenReinforcedNotification.objects.create_from_notification(
            notification
        )

    def test_partial_text_notification(self):
        """Test when the notification has enough to be considered a yaml dict but missing some fields"""

        notification = deepcopy(REINFORCE_NOTIFICATION)
        notification["text"] = (  # timestamp fields removed
            'aggressorAllianceName: <a href="showinfo:16159//99003214">Brave Collective</a>\naggressorCharacterID: 96914524\naggressorCorporationName: <a href="showinfo:2//98169165">Brave Newbies Inc.</a>\nitemID: &id001 1047210542765\nmercenaryDenShowInfoData:\n- showinfo\n- 85230\n- *id001\nplanetID: 40255101\nplanetShowInfoData:\n- showinfo\n- 11\n- 40255101\nsolarsystemID: 30004028\n'
        )

        self.assertRaises(
            ValueError,
            MercenaryDenReinforcedNotification.objects.create_from_notification,
            notification,
        )



from django.test import TestCase
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from . import test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        test_data_spec = [
            ModelSpec(
                "EvePlanet",
                ids=[
                    40255090,
                    40255092,
                    40255094,
                    40255096,
                    40255099,
                    40255101,
                    40255105,
                    40255125,
                ],
            ),
            ModelSpec(
                "EveType",
                ids=[
                    85230,
                ],
            ),
        ]
        create_testdata(test_data_spec, test_data_filename())



from django.core.management import BaseCommand

from allianceauth.services.hooks import get_extension_logger

from dens.esi import get_character_assets_from_esi
from dens.models import DenOwner

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    """Debugging command returning a character assets"""

    help = "Returns all the character assets from the ESI"

    def add_arguments(self, parser):
        parser.add_argument("character_id", type=int)

    def handle(self, *args, **options):
        character_id = options["character_id"]
        print(
            f"Trying to retrieve all ESI assets for user id {character_id}",
        )
        owner = DenOwner.objects.get(
            character_ownership__character__character_id=character_id
        )

        assets = get_character_assets_from_esi(owner)
        print("USER ASSETS ================================================")
        print(assets)
        print("============================================================")



"""Views."""

from django_datatables_view.base_datatable_view import BaseDatatableView

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.mixins import LoginRequiredMixin, PermissionRequiredMixin
from django.db import models
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from esi.decorators import token_required

from allianceauth.eveonline.evelinks import dotlan
from app_utils.views import link_html

from dens import tasks
from dens.models import ESI_SCOPES, DenOwner, MercenaryDen


@login_required
@permission_required("dens.basic_access")
def index(request):
    """Render index view."""
    return redirect("dens:dens")


@login_required
@permission_required("dens.basic_access")
def dens(request):
    """Render dens databatable"""
    return render(request, "dens/dens.html")


@login_required
@permission_required("dens.basic_access")
@token_required(scopes=ESI_SCOPES)
def add_owner(request, token):
    """View to add an owner"""
    character_ownership = get_object_or_404(
        request.user.character_ownerships.select_related("character"),
        character__character_id=token.character_id,
    )

    owner, created = DenOwner.objects.get_or_create(
        character_ownership=character_ownership
    )

    if not created:
        owner.enable()
        messages.success(
            request,
            _(
                "Successfully enabled owner %(owner_name)s. Starting to fetch mercenary dens."
            )
            % {"owner_name": owner.character_name},
        )
    else:
        messages.success(
            request,
            _(
                "Successfully created owner %(owner_name)s. Starting to fetch mercenary dens."
            )
            % {"owner_name": owner.character_name},
        )

    tasks.update_owner_dens.delay(owner.id)

    return redirect("dens:index")


# pylint: disable = too-many-ancestors
class MercenaryDensListJson(
    PermissionRequiredMixin, LoginRequiredMixin, BaseDatatableView
):
    """
    Datatable view rendering users mercenary dens
    """

    model = MercenaryDen
    permission_required = "dens.basic_access"
    columns = [
        "id",
        "owner_character_name",
        "owner_main_character_name",
        "planet_name",
        "solar_system_link",
        "location_html",
        "is_reinforced",
        "reinforcement_time",
        "region_name",
        "constellation_name",
        "solar_system_name",
    ]
    order_columns = [
        "pk",
    ]

    # pylint: disable = too-many-return-statements, inconsistent-return-statements
    def render_column(self, row: MercenaryDen, column):
        if column == "id":
            return row.pk

        if column == "owner_character_name":
            return row.owner.character_name

        if column == "owner_main_character_name":
            return (
                row.owner.character_ownership.user.profile.main_character.character_name
            )

        if column == "is_reinforced":
            return row.is_reinforced

        if column == "reinforcement_time":
            if reinforcement_time := row.reinforcement_time:
                return reinforcement_time.isoformat()
            return

        if result := self._render_location(row, column):
            return result

    def get_initial_queryset(self):
        user = self.request.user
        if user.has_perm("dens.manager"):
            den_query = MercenaryDen.objects.all()
        elif user.has_perm("dens.alliance_view"):
            alliance = self.request.user.profile.main_character.alliance
            den_query = MercenaryDen.objects.filter_alliance_dens(alliance)
        elif user.has_perm("dens.corporation_view"):
            corporation = self.request.user.profile.main_character.corporation
            den_query = MercenaryDen.objects.filter_corporation_dens(corporation)
        else:
            den_query = MercenaryDen.objects.filter_user_dens(user)
        return den_query

    def filter_queryset(self, qs):
        qs = self._apply_search_filter(qs, 0, "owner")
        qs = self._apply_search_filter(
            qs, 1, "owner__character_ownership__user__profile__main__character_name"
        )

        if search := self.request.GET.get("search[value]", None):
            # TODO improve search
            qs = qs.filter(location__name__istartswith=search)

        return qs

    # pylint: disable = too-many-return-statements
    def _render_location(self, row: MercenaryDen, column):
        """Renders location for mercenary den dataview display"""
        planet = row.location
        solar_system = planet.eve_solar_system
        constellation = solar_system.eve_constellation
        region = constellation.eve_region

        if column == "planet_name":
            return planet.name

        if column == "solar_system_name":
            return solar_system.name

        if column == "region_name":
            return region.name

        if column == "constellation_name":
            return constellation.name

        if column == "solar_system_link":
            solar_system_link = format_html(
                '{}&nbsp;<span class="text-null-sec">{}</span>',
                link_html(
                    dotlan.solar_system_url(solar_system.name), solar_system.name
                ),
                round(solar_system.security_status, 1),
            )
            return solar_system_link

        if column == "location_html":
            location_html = format_html(
                "{}<br><em>{}</em>", constellation.name, region.name
            )
            return location_html

        return None

    def _apply_search_filter(self, qs, column_num, field) -> models.QuerySet:
        my_filter = self.request.GET.get(f"columns[{column_num}][search][value]", None)
        if my_filter:
            if self.request.GET.get(f"columns[{column_num}][search][regex]", False):
                kwargs = {f"{field}__iregex": my_filter}
            else:
                kwargs = {f"{field}__istartswith": my_filter}
            return qs.filter(**kwargs)
        return qs


def dens_fdd_data(request) -> JsonResponse:
    """List for the drop-down fields"""
    qs = MercenaryDensListJson.get_initial_queryset(request)
    columns = request.GET.get("columns")
    result = {}
    if columns:
        for column in columns.split(","):
            options = _calc_options(request, qs, column)
            result[column] = sorted(list(set(options)), key=str.casefold)
    return JsonResponse(result, safe=False)


def _calc_options(request, qs, column):
    if column == "owner_character_name":
        return qs.values_list(
            "owner__character_ownership__character__character_name",
            flat=True,
        )

    if column == "owner_main_character_name":
        return qs.values_list(
            "owner__character_ownership__user__profile__main__character_name",
            flat=True,
        )

    return [f"** ERROR: Invalid column name '{column}' **"]



from django.test import TestCase

from dens import tasks
from dens.esi import MissingTokenError
from dens.models import DenOwner, MercenaryDen
from dens.tests.testdata.load_eveuniverse import load_eveuniverse
from dens.tests.utils import create_fake_den_owners, create_fake_dens_for_owner


class TestNotification(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

        owners = create_fake_den_owners(2)

        create_fake_dens_for_owner(owners[0], [40255090, 40255092])
        create_fake_dens_for_owner(owners[1], [40255094])

    def test_enabled_owners(self):
        all_owners = DenOwner.objects.enabled_owners()
        self.assertEqual(2, all_owners.count())
        all_owners = list(all_owners)

        all_owners[0].disable()

        enabled_owners = DenOwner.objects.enabled_owners()
        self.assertEqual(1, enabled_owners.count())
        self.assertIn(all_owners[1], enabled_owners)

    def test_owner_gets_disabled(self):
        """Tests all functions that can disable owners to ensure they get disabled and MissingToken is raised"""

        owner = DenOwner.objects.all()[0]

        self.assertRaises(MissingTokenError, tasks.update_owner_dens, owner.id)
        owner = DenOwner.objects.get(id=owner.id)
        self.assertFalse(owner.is_enabled)

        owner.enable()
        self.assertTrue(owner.is_enabled)

        self.assertRaises(
            MissingTokenError, tasks.create_mercenary_den, owner.id, {"item_id": 123}
        )
        owner = DenOwner.objects.get(id=owner.id)
        self.assertFalse(owner.is_enabled)

        owner.enable()
        self.assertTrue(owner.is_enabled)

        self.assertRaises(MissingTokenError, tasks.update_owner_notifications, owner.id)
        owner = DenOwner.objects.get(id=owner.id)
        self.assertFalse(owner.is_enabled)

    def test_new_den_on_disabled_owner_planet(self):
        """Checks what happens when a new den is created on a planet with a disabled owner's den"""
        EVE_PLANET = 40255090

        owners = list(DenOwner.objects.all())

        owners[0].disable()

        self.assertEqual(2, MercenaryDen.objects.filter(owner=owners[0]).count())
        self.assertTrue(
            MercenaryDen.objects.filter(
                location_id=EVE_PLANET, owner=owners[0]
            ).exists()
        )

        create_fake_dens_for_owner(owners[1], [EVE_PLANET])

        self.assertEqual(1, MercenaryDen.objects.filter(location=EVE_PLANET).count())



"""
Handles interactions with the structuretimers application

https://gitlab.com/ErikKalkoken/aa-structuretimers
"""

from structuretimers.models import Timer

from eveuniverse.models import EveType

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo

from dens.models import MercenaryDenReinforcedNotification


def create_structuretimer_timer(notification: MercenaryDenReinforcedNotification):
    """
    Creates a timer for this reinforcement in the structruretimer app
    """

    eve_character = notification.den.owner.character_ownership.character
    try:
        eve_corp = eve_character.corporation
    except EveCorporationInfo.DoesNotExist:
        eve_corp = EveCorporationInfo.objects.create_corporation(
            notification.den.owner.character_ownership.character.corporation_id
        )

    try:
        eve_alliance = eve_corp.alliance
    except EveAllianceInfo.DoesNotExist:
        eve_alliance = EveAllianceInfo.objects.create_alliance(eve_corp.alliance_id)

    structure_type, _ = EveType.objects.get_or_create_esi(id=85230)

    Timer.objects.create(
        date=notification.exit_reinforcement,
        details_notes=f"Reinforced by {notification.reinforced_by.character_name}",
        eve_alliance=eve_alliance,
        eve_character=eve_character,
        eve_corporation=eve_corp,
        eve_solar_system=notification.den.location.eve_solar_system,
        location_details=notification.den.location.name,
        objective=Timer.Objective.FRIENDLY,
        owner_name=eve_character.character_name,
        structure_type=structure_type,
        structure_name=notification.den.location.name.split()[-1],
        timer_type=Timer.Type.FINAL,
    )



"""Managers."""

import re
from typing import Optional

import yaml

from django.contrib.auth.models import User
from django.db import models
from django.utils import timezone
from eveuniverse.models import EvePlanet

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger

from dens.esi import get_time_eve

from . import models as den_models

logger = get_extension_logger(__name__)


class DenOwnerManager(models.Manager):
    """Manager for DenOwner"""

    def enabled_owners(self):
        """Returns all active den owners"""
        return self.filter(is_enabled=True)


class MercenaryDenManager(models.Manager):
    """Manager for mercenary dens"""

    def get_owner_dens_ids_set(self, owner: "den_models.DenOwner") -> set[int]:
        """Returns a set with the id of all dens anchored by this owner"""
        return set(self.filter(owner=owner).values_list("id", flat=True))

    def create(
        self, owner: "den_models.DenOwner", den_id: int, location: EvePlanet, **kwargs
    ):
        """Creates a mercenary den and returns it"""
        if self.filter(location=location).exists():
            # Cleans a previous den that was probably the leftover of a disabled owner
            self.get(location=location).delete()
        return super().create(owner=owner, id=den_id, location=location, **kwargs)

    def filter_alliance_dens(self, alliance: EveAllianceInfo):
        """Returns all mercenary dens from owners in the given alliance"""
        return self.filter(
            owner__character_ownership__character__alliance_id=alliance.alliance_id
        )

    def filter_corporation_dens(self, corporation: EveCorporationInfo):
        """Returns all mercenary dens from owners in a given corporation"""
        return self.filter(
            owner__character_ownership__character__corporation_id=corporation.corporation_id
        )

    def filter_user_dens(self, user: User):
        """Return all mercenary dens associated with a user"""
        return self.filter(owner__character_ownership__user=user)


class MercernaryDenReinforcedNotificationManager(models.Manager):
    """Manager for MercenaryDenReinforcedNotification"""

    def is_notification_id_known(self, notification_id: int) -> bool:
        """Will check if the notification id is in the database"""
        return self.filter(id=notification_id).exists()

    def parse_information_from_notification(self, notification_dict: dict) -> dict:
        """Generates and saves a new notification from the notification dict of the ESI"""

        text = yaml.load(notification_dict["text"], Loader=yaml.Loader)

        if not isinstance(text, dict):  # input string wasn't formatted correctly
            logger.error(
                "Couldn't make a notification out of %s the format doesn't seem like valid yaml",
                notification_dict,
            )
            raise ValueError(
                f"Couldn't make a notification out of {notification_dict} the format doesn't seem like valid yaml"
            )

        alliance_match = re.match(
            r"<a href=\"showinfo:16159//(?P<alliance_id>.+)\">(?P<alliance_name>.+)</a>",
            text["aggressorAllianceName"],
        )
        corporation_match = re.match(
            r"<a href=\"showinfo:2//(?P<corporation_id>.+)\">(?P<corporation_name>.+)</a>",
            text["aggressorCorporationName"],
        )

        try:
            return {
                "aggressor_character_id": text["aggressorCharacterID"],
                "planet_id": text["planetShowInfoData"][2],
                "timestamp_entered": text["timestampEntered"],
                "timestamp_exited": text["timestampExited"],
                "aggressor_alliance": (
                    {
                        "name": alliance_match.group("alliance_name"),
                        "id": alliance_match.group("alliance_id"),
                    }
                    if alliance_match
                    else None
                ),
                "aggressor_corporation": {
                    "name": corporation_match.group("corporation_name"),
                    "id": corporation_match.group("corporation_id"),
                },
                "solar_system_id": text["solarsystemID"],
                "mercenary_den_id": text["mercenaryDenShowInfoData"],
            }
        except KeyError as e:  # Missing fields
            logger.error(
                "Couldn't make a notification out of %s some field is missing: %s",
                notification_dict,
                e,
            )
            raise ValueError(
                f"Couldn't make a notification out of {notification_dict} some field is missing: {e}"
            ) from e

    def create_from_notification(
        self, notification: dict
    ) -> Optional["den_models.MercenaryDenReinforcedNotification"]:
        """
        Creates a den reinforced notification from an ESI notification.
        """

        notification_information = self.parse_information_from_notification(
            notification
        )
        if notification_information is None:
            logger.error("Couldn't make a notification out of %s", notification)
            raise ValueError(f"Couldn't make a notification out of {notification}")

        associated_planet, _ = EvePlanet.objects.get_or_create_esi(
            id=notification_information["planet_id"]
        )

        eve_character_id = notification_information["aggressor_character_id"]
        reinforced_by = EveCharacter.objects.get_character_by_id(eve_character_id)
        if reinforced_by is None:
            reinforced_by = EveCharacter.objects.create_character(eve_character_id)
        logger.debug("Reinforced by %s", reinforced_by)
        entered_reinforce = get_time_eve(notification_information["timestamp_entered"])
        exited_reinforce = get_time_eve(notification_information["timestamp_exited"])

        try:
            associated_den = den_models.MercenaryDen.objects.get(
                location=associated_planet
            )
        except den_models.MercenaryDen.DoesNotExist:
            logger.info(
                "Trying to parse the notification of a non existing den on planet id %s",
                associated_planet.id,
            )
            if (
                exited_reinforce > timezone.now()
            ):  # Future notification of unknown den, will need to reparse it
                return None
            associated_den = None  # Past notification, can be safely ignored

        notification = self.create(
            id=notification["notification_id"],
            den=associated_den,
            reinforced_by=reinforced_by,
            enter_reinforcement=entered_reinforce,
            exit_reinforcement=exited_reinforce,
        )

        return notification



"""Esi related functions"""

import datetime

from django.utils import timezone
from esi.clients import EsiClientProvider
from esi.models import Token

from allianceauth.services.hooks import get_extension_logger

# from .models import DenOwner
from . import __version__, models

MERCENARY_DENS_TYPE_IDS = [
    85230,
    85980,
]
TEMPERATE_PLANET_TYPE_ID = 11


class MissingTokenError(Exception):
    """Raised when the token for this user isn't found"""


esi = EsiClientProvider(app_info_text=f"aa-dens v{__version__}")
logger = get_extension_logger(__name__)


def get_character_assets_from_esi(owner: "models.DenOwner") -> list[dict]:
    """Returns all character assets from the ESI"""
    logger.info("Fetching esi asset for user id %s", owner.id)

    try:
        assets = esi.client.Assets.get_characters_character_id_assets(
            character_id=owner.character_id,
            token=owner.fetch_token().valid_access_token(),
        ).results()
    except Token.DoesNotExist as e:
        logger.error("Missing ESI token for owner %s")
        raise MissingTokenError("Missing ESI token when fetching owner assets") from e

    logger.debug(assets)

    return assets


def get_owner_anchored_dens_from_esi(owner: "models.DenOwner") -> list[dict]:
    """Return all dens locations from the ESI"""

    den_assets = [
        asset
        for asset in get_character_assets_from_esi(owner)
        if asset["type_id"] in MERCENARY_DENS_TYPE_IDS
        and asset["location_type"] == "solar_system"
    ]

    return den_assets


def get_esi_asset_location(owner: "models.DenOwner", item_id: int) -> (int, int, int):
    """Returns the position of an item"""
    logger.info(
        "Fetching item id %s location from owner %s esi information", item_id, owner
    )

    try:
        position = esi.client.Assets.post_characters_character_id_assets_locations(
            character_id=owner.character_id,
            item_ids=[item_id],
            token=owner.fetch_token().valid_access_token(),
        ).result()[0]["position"]
    except Token.DoesNotExist as e:
        logger.error("Missing ESI token for owner %s")
        raise MissingTokenError(
            "Missing ESI token when fetching owner asset location"
        ) from e

    logger.debug(position)

    return position["x"], position["y"], position["z"]


def get_owner_notifications(owner: "models.DenOwner") -> list[dict]:
    """Returns notifications from an owner"""
    logger.info("Fetching notifications for den owner id %s", owner.id)

    try:
        notifications = esi.client.Character.get_characters_character_id_notifications(
            character_id=owner.character_id,
            token=owner.fetch_token().valid_access_token(),
        ).results()
    except Token.DoesNotExist as e:
        logger.error("Missing ESI token for owner %s")
        raise MissingTokenError(
            "Missing ESI token when fetching owner notifications"
        ) from e

    logger.debug(notifications)

    return notifications


def get_owner_mercenarydenreinforced_notifications_from_esi(
    owner: "models.DenOwner",
) -> list[dict]:
    """Returns only the `MercenaryDenReinforced` notifications from this owner"""

    return [
        notification
        for notification in get_owner_notifications(owner)
        if notification["type"] == "MercenaryDenReinforced"
    ]


def get_time_eve(dt: int) -> datetime.datetime:
    """
    Formula to parse ESI timestamps to datetime
    https://forums.eveonline.com/t/timestamp-format-in-notifications-by-esi/230395
    """
    microseconds = dt / 10
    seconds, microseconds = divmod(microseconds, 1000000)
    days, seconds = divmod(seconds, 86400)
    return datetime.datetime(1601, 1, 1, tzinfo=timezone.utc) + datetime.timedelta(
        days, seconds
    )



from django.core.management import BaseCommand

from allianceauth.services.hooks import get_extension_logger

from dens import tasks

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    help = "Update all den owners"

    def handle(self, *args, **kwargs):
        logger.info("Initializing update of den owners from command line")
        tasks.update_all_den_owners.delay()



# Generated by Django 4.2.16 on 2025-01-26 22:33

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("authentication", "0024_alter_userprofile_language"),
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app"),
                    (
                        "corporation_view",
                        "Can view all dens anchored by members of their corporation",
                    ),
                    (
                        "alliance_view",
                        "Can view all dens anchored by members of their alliance",
                    ),
                    ("manager", "Can see all user's mercenary dens"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="DenOwner",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="Disabled characters won't be synced",
                    ),
                ),
                (
                    "character_ownership",
                    models.ForeignKey(
                        help_text="Character used to sync mercenary dens",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MercenaryDen",
            fields=[
                (
                    "id",
                    models.BigIntegerField(
                        help_text="Eve online id of the den",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "location",
                    models.ForeignKey(
                        help_text="Location of the den",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.eveplanet",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Character that anchored the den",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="dens.denowner",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MercenaryDenReinforcedNotification",
            fields=[
                ("id", models.BigIntegerField(primary_key=True, serialize=False)),
                (
                    "enter_reinforcement",
                    models.DateTimeField(help_text="Timer when the den was reinforced"),
                ),
                (
                    "exit_reinforcement",
                    models.DateTimeField(
                        help_text="Timer when the den will leave reinforcement"
                    ),
                ),
                (
                    "den",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="dens.mercenaryden",
                    ),
                ),
                (
                    "reinforced_by",
                    models.ForeignKey(
                        help_text="Character that reinforced the Mercenary Den",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveonline.evecharacter",
                    ),
                ),
            ],
        ),
    ]



"""Admin site."""

from django.contrib import admin
from django.db.models import QuerySet

from dens.models import DenOwner, MercenaryDen, MercenaryDenReinforcedNotification


class DenOwnerUserFilter(admin.SimpleListFilter):
    """Filters the DenOwners by their user"""

    title = "user"
    parameter_name = "user"

    def lookups(self, request, model_admin):
        return DenOwner.objects.values_list(
            "character_ownership__user__id", "character_ownership__user__username"
        ).distinct()

    def queryset(self, request, queryset: QuerySet[DenOwner]):
        if value := self.value():
            return queryset.filter(character_ownership__user__id=value)
        else:
            return queryset


@admin.register(DenOwner)
class DenOwnerAdmin(admin.ModelAdmin):
    list_display = ["character_name", "user_name", "dens_count", "is_enabled"]
    readonly_fields = ["character_ownership"]
    list_filter = ["is_enabled", DenOwnerUserFilter]

    def has_add_permission(self, request):
        return False

    @admin.display(description="#Dens anchored under this owner")
    def dens_count(self, owner: DenOwner):
        return len(MercenaryDen.objects.get_owner_dens_ids_set(owner))


class DenUserFilter(admin.SimpleListFilter):
    """Filters the MercenaryDens by their user"""

    title = "user"
    parameter_name = "user"

    def lookups(self, request, model_admin):
        return MercenaryDen.objects.values_list(
            "owner__character_ownership__user__id",
            "owner__character_ownership__user__username",
        ).distinct()

    def queryset(self, request, queryset: QuerySet[DenOwner]):
        if value := self.value():
            return queryset.filter(owner__character_ownership__user__id=value)
        else:
            return queryset


@admin.register(MercenaryDen)
class MercenaryDenAdmin(admin.ModelAdmin):
    list_display = ["location", "owner", "is_reinforced", "reinforcement_time"]
    list_filter = ["owner", DenUserFilter]

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=...):
        return False


@admin.register(MercenaryDenReinforcedNotification)
class MercenaryDenReinforcedNotificationAdmin(admin.ModelAdmin):
    list_display = ["den", "reinforced_by", "enter_reinforcement", "exit_reinforcement"]
    search_fields = [
        "reinforced_by__character_name",
        "reinforced_by__corporation_name",
        "reinforced_by__alliance_name",
    ]

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=...):
        return False

    def has_change_permission(self, request, obj=...):
        return False



"""
Handles interactions with the timerboard application

https://allianceauth.readthedocs.io/en/v4.6.1/features/apps/timerboard.html
"""

from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.timerboard.models import Timer

from dens.models import MercenaryDenReinforcedNotification


def create_timerboard_timer(notification: MercenaryDenReinforcedNotification):
    """
    Creates a timer for this reinforcement in the timerboard app
    """

    try:
        eve_corp = notification.den.owner.character_ownership.character.corporation
    except EveCorporationInfo.DoesNotExist:
        eve_corp = EveCorporationInfo.objects.create_corporation(
            notification.den.owner.character_ownership.character.corporation_id
        )

    Timer.objects.create(
        details=f"Mercenary den reinforced by {notification.reinforced_by.character_name}",
        system=notification.den.location.eve_solar_system.name,
        planet_moon=notification.den.location.name,
        structure=Timer.Structure.MERCDEN,
        timer_type=Timer.TimerType.FINAL,
        objective=Timer.Objective.FRIENDLY,
        eve_time=notification.exit_reinforcement,
        eve_corp=eve_corp,
    )



import datetime

from dateutil.tz import tzutc

from eveuniverse.models import EvePlanet

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo
from app_utils.testing import create_fake_user

from dens.models import DenOwner, MercenaryDen, MercenaryDenReinforcedNotification
from dens.tasks import create_reinforce_notification

REINFORCE_NOTIFICATION = {
    "is_read": True,
    "notification_id": 2064665856,
    "sender_id": 1000438,
    "sender_type": "corporation",
    "text": 'aggressorAllianceName: <a href="showinfo:16159//99003214">Brave Collective</a>\naggressorCharacterID: 96914524\naggressorCorporationName: <a href="showinfo:2//98169165">Brave Newbies Inc.</a>\nitemID: &id001 1047210542765\nmercenaryDenShowInfoData:\n- showinfo\n- 85230\n- *id001\nplanetID: 40255101\nplanetShowInfoData:\n- showinfo\n- 11\n- 40255101\nsolarsystemID: 30004028\ntimestampEntered: 133761583913385305\ntimestampExited: 133762405913385305\ntypeID: 85230\n',
    "timestamp": datetime.datetime(2024, 11, 15, 15, 33, tzinfo=tzutc()),
    "type": "MercenaryDenReinforced",
}

NO_ALLIANCE_REINFORCE_NOTIFICATION = {
    "is_read": True,
    "notification_id": 2064898731,
    "sender_id": 1000438,
    "sender_type": "corporation",
    "text": 'aggressorAllianceName: Unknown\naggressorCharacterID: 185018643\naggressorCorporationName: <a href="showinfo:2//828783563">exiles.</a>\nitemID: &id001 1047212774639\nmercenaryDenShowInfoData:\n- showinfo\n- 85230\n- *id001\nplanetID: 40255101\nplanetShowInfoData:\n- showinfo\n- 11\n- 40256161\nsolarsystemID: 30004045\ntimestampEntered: 133761878739722751\ntimestampExited: 133762846589722751\ntypeID: 85230\n',
    "timestamp": datetime.datetime(2024, 11, 15, 23, 45, tzinfo=datetime.timezone.utc),
    "type": "MercenaryDenReinforced",
}


def create_fake_den_owner() -> DenOwner:
    """Returns a fake den owner for test purposes"""

    user = create_fake_user(
        character_name="Ji'had Rokym",
        character_id=2119057263,
        corporation_id=98609038,
        corporation_ticker="ROKYM",
        corporation_name="Rokym's managment organisation",
        alliance_name="Tocards Professionnels",
        alliance_id=99011247,
    )

    char_ownership, _ = CharacterOwnership.objects.get_or_create(
        character=user.profile.main_character, user=user, owner_hash="fake_hash"
    )

    owner = DenOwner.objects.create(character_ownership=char_ownership)

    return owner


def create_fake_den_owner_no_alliance() -> DenOwner:
    """Returns a fake den owner not in an alliance"""

    user = create_fake_user(
        character_name="Ji'had Rokym",
        character_id=2119057263,
        corporation_id=98609038,
        corporation_ticker="ROKYM",
        corporation_name="Rokym's managment organisation",
    )

    char_ownership, _ = CharacterOwnership.objects.get_or_create(
        character=user.profile.main_character, user=user, owner_hash="fake_hash"
    )

    owner = DenOwner.objects.create(character_ownership=char_ownership)

    return owner


def create_fake_den_owners(number: int) -> list[DenOwner]:
    """Create as many den owners as needed all in their individual corporations and alliances"""

    owners = []
    for i in range(number):
        user = create_fake_user(
            character_name=f"User {i}",
            character_id=1000 + i,
            corporation_name=f"Corporation {i}",
            corporation_ticker=f"CORP{i}",
            corporation_id=2000 + i,
            alliance_name=f"Alliance {i}",
            alliance_id=3000 + i,
        )

        eve_alliance = EveAllianceInfo.objects.create(
            alliance_id=3000 + i,
            alliance_name=f"Alliance {i}",
            alliance_ticker=f"ALLY{i}",
            executor_corp_id=2000 + i,
        )
        EveCorporationInfo.objects.create(
            corporation_id=2000 + i,
            corporation_name=f"Corporation{i}",
            corporation_ticker=f"CORP{i}",
            member_count=1,
            alliance=eve_alliance,
        )

        char_ownership, _ = CharacterOwnership.objects.get_or_create(
            character=user.profile.main_character, user=user, owner_hash=f"fake_hash{i}"
        )

        owners.append(DenOwner.objects.create(character_ownership=char_ownership))

    return owners


def create_fake_den(owner: DenOwner) -> MercenaryDen:
    """Create a fake den belonging to the passed owner for test purposes"""

    planet_id = 40255101
    planet = EvePlanet.objects.get(id=planet_id)

    den = MercenaryDen.objects.create(owner=owner, den_id=1, location=planet)

    return den


def create_fake_dens_for_owner(
    owner: DenOwner, planet_ids: list[int]
) -> list[MercenaryDen]:
    """Create mercenary dens on the selected planets for the given owner"""
    existing_dens = MercenaryDen.objects.count()
    dens = []

    for i in range(len(planet_ids)):
        planet_id = planet_ids[i]
        planet = EvePlanet.objects.get(id=planet_id)
        dens.append(
            MercenaryDen.objects.create(
                owner=owner, location=planet, den_id=existing_dens + i
            )
        )

    return dens


def create_fake_notification():
    """Creates user, den and stores a den reinforcement notification"""
    owner = create_fake_den_owner()
    create_fake_den(owner)

    create_reinforce_notification(REINFORCE_NOTIFICATION)

    return MercenaryDenReinforcedNotification.objects.all()[0]



from unittest.mock import call, patch

from django.test import TestCase

from dens.esi import MissingTokenError
from dens.models import DenOwner
from dens.tasks import update_all_den_owners
from dens.tests.utils import create_fake_den_owners


class TestTasks(TestCase):
    """Tests for different tasks"""

    @patch("dens.tasks.get_owner_anchored_dens_from_esi")
    def test_owner_updates_after_error(self, get_owner_anchored_dens_from_esi_mock):
        """If an owner errors out because of a missing token the next owner should still be updated"""
        get_owner_anchored_dens_from_esi_mock.side_effect = [
            MissingTokenError(),
            [],
        ]

        owners = create_fake_den_owners(2)

        update_all_den_owners()

        self.assertEqual(get_owner_anchored_dens_from_esi_mock.call_count, 2)
        calls = [call(owners[0]), call(owners[1])]
        get_owner_anchored_dens_from_esi_mock.assert_has_calls(calls, any_order=True)

        self.assertEqual(DenOwner.objects.filter(is_enabled=False).count(), 1)



# Generated by Django 4.2.16 on 2025-02-19 10:45

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("dens", "0002_alter_mercenarydenreinforcednotification_den"),
    ]

    operations = [
        migrations.AlterField(
            model_name="mercenaryden",
            name="location",
            field=models.OneToOneField(
                help_text="Location of the den",
                on_delete=django.db.models.deletion.CASCADE,
                to="eveuniverse.eveplanet",
            ),
        ),
    ]



"""Routes."""

from django.urls import path

from . import views

app_name = "dens"

urlpatterns = [
    path("", views.index, name="index"),
    path("add_owner", views.add_owner, name="add_owner"),
    path("dens", views.dens, name="dens"),
    path("dens_data", views.MercenaryDensListJson.as_view(), name="dens_data"),
]



from django.test import TestCase

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo

from dens.models import DenOwner, MercenaryDen
from dens.tests.testdata.load_eveuniverse import load_eveuniverse
from dens.tests.utils import create_fake_den_owners, create_fake_dens_for_owner


class TestNotification(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

        owners = create_fake_den_owners(2)

        create_fake_dens_for_owner(owners[0], [40255090, 40255092])
        create_fake_dens_for_owner(owners[1], [40255094])

    def test_pre_conditions(self):
        """Test case to check that the situations is as expected"""
        self.assertEqual(DenOwner.objects.count(), 2)
        self.assertEqual(MercenaryDen.objects.count(), 3)

        self.assertTrue(EveCorporationInfo.objects.filter(corporation_id=2000).exists())
        self.assertTrue(EveCorporationInfo.objects.filter(corporation_id=2001).exists())

        self.assertTrue(EveAllianceInfo.objects.filter(alliance_id=3000).exists())
        self.assertTrue(EveAllianceInfo.objects.filter(alliance_id=3001).exists())

    def test_get_all_dens(self):
        self.assertEqual(MercenaryDen.objects.all().count(), 3)

    def test_get_alliance_dens(self):
        alliance_1 = EveAllianceInfo.objects.get(alliance_id=3000)
        alliance_2 = EveAllianceInfo.objects.get(alliance_id=3001)

        self.assertEqual(
            MercenaryDen.objects.filter_alliance_dens(alliance_1).count(), 2
        )
        self.assertEqual(
            MercenaryDen.objects.filter_alliance_dens(alliance_2).count(), 1
        )

    def test_get_corporation_dens(self):
        corporation_1 = EveCorporationInfo.objects.get(corporation_id=2000)
        corporation_2 = EveCorporationInfo.objects.get(corporation_id=2001)

        self.assertEqual(
            MercenaryDen.objects.filter_corporation_dens(corporation_1).count(), 2
        )
        self.assertEqual(
            MercenaryDen.objects.filter_corporation_dens(corporation_2).count(), 1
        )



"""Hooks for charlink application"""

from charlink.app_imports.utils import (  # pylint: disable=import-error
    AppImport,
    LoginImport,
)

from django.contrib import messages
from django.contrib.auth.models import Permission, User
from django.db.models import Exists, OuterRef
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCharacter
from app_utils.django import users_with_permission

from dens import tasks
from dens.models import ESI_SCOPES, DenOwner


# pylint: disable=duplicate-code
def _add_character(request, token):
    character_ownership = get_object_or_404(
        request.user.character_ownerships.select_related("character"),
        character__character_id=token.character_id,
    )

    owner, created = DenOwner.objects.get_or_create(
        character_ownership=character_ownership
    )

    if not created:
        owner.enable()
        messages.success(
            request,
            _(
                "Successfully enabled owner %(owner_name)s. Starting to fetch mercenary dens."
            )
            % {"owner_name": owner.character_name},
        )
    else:
        messages.success(
            request,
            _(
                "Successfully created owner %(owner_name)s. Starting to fetch mercenary dens."
            )
            % {"owner_name": owner.character_name},
        )

    tasks.update_owner_dens.delay(owner.id)


def _check_permissions(user: User) -> bool:
    return user.has_perm("dens.basic_access")


def _is_character_added(character: EveCharacter) -> bool:
    return DenOwner.objects.filter(character_ownership__character=character).exists()


def _users_with_perms():
    return users_with_permission(
        Permission.objects.get(content_type__app_label="dens", codename="basic_access")
    )


app_import = AppImport(
    "dens",
    [
        LoginImport(
            app_label="dens",
            unique_id="default",
            field_label=_("Mercenary Dens"),
            add_character=_add_character,
            scopes=ESI_SCOPES,
            check_permissions=_check_permissions,
            is_character_added=_is_character_added,
            is_character_added_annotation=Exists(
                DenOwner.objects.filter(
                    character_ownership__character_id=OuterRef("pk")
                )
            ),
            get_users_with_perms=_users_with_perms,
        )
    ],
)



# Generated by Django 4.2.16 on 2025-01-30 11:50

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("dens", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="mercenarydenreinforcednotification",
            name="den",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="dens.mercenaryden",
            ),
        ),
    ]



"""Models."""

import datetime
from typing import Optional

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from esi.models import Token
from eveuniverse.models import EvePlanet

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

from dens.managers import (
    DenOwnerManager,
    MercenaryDenManager,
    MercernaryDenReinforcedNotificationManager,
)

ESI_SCOPES = [
    "esi-assets.read_assets.v1",
    "esi-characters.read_notifications.v1",
]

logger = get_extension_logger(__name__)


class General(models.Model):
    """A meta model for app permissions."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            (
                "corporation_view",
                "Can view all dens anchored by members of their corporation",
            ),
            (
                "alliance_view",
                "Can view all dens anchored by members of their alliance",
            ),
            ("manager", "Can see all user's mercenary dens"),
        )


class DenOwner(models.Model):
    """Represents a character that will drop mercenary dens"""

    objects = DenOwnerManager()

    character_ownership = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.CASCADE,
        related_name="+",
        help_text="Character used to sync mercenary dens",
    )

    is_enabled = models.BooleanField(
        default=True, db_index=True, help_text="Disabled characters won't be synced"
    )

    def __str__(self):
        return self.character_name

    def fetch_token(self) -> Token:
        """Return valid token for this mining corp or raise exception on any error."""
        if not self.character_ownership:
            raise RuntimeError("This owner has no character configured.")
        token = (
            Token.objects.filter(
                character_id=self.character_ownership.character.character_id
            )
            .require_scopes(ESI_SCOPES)
            .require_valid()
            .first()
        )
        if not token:
            raise Token.DoesNotExist(f"{self}: No valid token found.")
        return token

    @property
    def character_name(self) -> str:
        """Returns the character name"""
        return self.character_ownership.character.character_name

    @property
    def user_name(self) -> str:
        """Returns the user name"""
        return self.character_ownership.user.username

    @property
    def character_id(self) -> int:
        """Returns the character id"""
        return self.character_ownership.character.character_id

    def enable(self):
        """Sets an owner back to activity"""
        self.is_enabled = True
        self.save()

    def disable(self):
        """Disables an owner"""
        self.is_enabled = False
        self.save()


class MercenaryDen(models.Model):
    """Represents anchored mercenary dens"""

    objects = MercenaryDenManager()

    id = models.BigIntegerField(
        primary_key=True, help_text=_("Eve online id of the den")
    )

    owner = models.ForeignKey(
        DenOwner,
        on_delete=models.CASCADE,
        help_text=_("Character that anchored the den"),
    )
    location = models.OneToOneField(
        EvePlanet, on_delete=models.CASCADE, help_text=_("Location of the den")
    )

    def __str__(self) -> str:
        return f"Den {self.location.name}"

    @property
    def is_reinforced(self) -> bool:
        """True if there's an unexited reinforcement notification"""
        now = timezone.now()
        return MercenaryDenReinforcedNotification.objects.filter(
            den=self, exit_reinforcement__gt=now
        ).exists()

    @property
    def reinforcement_time(self) -> Optional[datetime.datetime]:
        """Return the den reinforcement time if it exists"""
        now = timezone.now()
        try:
            notification = MercenaryDenReinforcedNotification.objects.get(
                den=self, exit_reinforcement__gt=now
            )
            return notification.exit_reinforcement
        except MercenaryDenReinforcedNotification.DoesNotExist:
            pass
        return None


class MercenaryDenReinforcedNotification(models.Model):
    """Represents the notification of an owner den reinforced"""

    objects = MercernaryDenReinforcedNotificationManager()

    id = models.BigIntegerField(primary_key=True)

    den = models.ForeignKey(MercenaryDen, on_delete=models.SET_NULL, null=True)

    reinforced_by = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE,
        help_text="Character that reinforced the Mercenary Den",
    )
    enter_reinforcement = models.DateTimeField(
        help_text=_("Timer when the den was reinforced")
    )
    exit_reinforcement = models.DateTimeField(
        help_text=_("Timer when the den will leave reinforcement")
    )

    def __str__(self) -> str:
        return f"Den {self.den.location.name} reinforced by {self.reinforced_by.character_name}"

    def is_in_future(self) -> bool:
        """True if the timer is in the future"""
        return self.exit_reinforcement > timezone.now()



"""
Interactions with the aa-discordbot application

https://github.com/Solar-Helix-Independent-Transport/allianceauth-discordbot
"""

from aadiscordbot.tasks import send_message
from discord import Embed

from dens.models import DenOwner, MercenaryDenReinforcedNotification


def basic_embed(title: str) -> Embed:
    """Returns a basic embed for the aa-den application"""

    e = Embed(
        title=title,
    )
    e.set_footer(
        text="aa-dens",
        icon_url="https://gitlab.com/uploads/-/system/project/avatar/64868536/tents_1_.png",
    )

    return e


def send_reinforced_notification_to_user(
    notification: MercenaryDenReinforcedNotification,
):
    """Sends a discord message a user warning that a mercenary den has been reinforced"""

    user = notification.den.owner.character_ownership.user

    e = basic_embed(
        title="Mercenary Den reinforced",
    )
    e.add_field(name="Location", value=notification.den.location.name)
    e.add_field(name="Reinforced by", value=notification.reinforced_by.character_name)
    e.add_field(
        name="Exit reinforcement", value=notification.exit_reinforcement.isoformat()
    )

    send_message(user=user, embed=e)


def send_disable_notification_to_user(owner: DenOwner):
    """Sends a discord message to a user warning that an owner has been disabled"""

    e = basic_embed(
        title="Mercenary Den owner disabled",
    )
    e.set_author(
        name=owner.character_name,
        icon_url=owner.character_ownership.character.portrait_url_64,
    )
    e.description = f"Your mercenary den owner {owner} is missing an ESI token and has been disabled"



"""App settings."""

from django.conf import settings

from app_utils.app_settings import clean_setting

EXAMPLE_SETTING_ONE = getattr(settings, "EXAMPLE_SETTING_ONE", None)

DENS_ADMIN_NOTIFICATIONS_ENABLED = clean_setting(
    "DENS_ADMIN_NOTIFICATIONS_ENABLED", True
)



"""Tasks."""

from celery import group, shared_task

from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from eveuniverse.constants import EveGroupId
from eveuniverse.core.evesdeapi import nearest_celestial
from eveuniverse.models import EvePlanet

import app_utils.allianceauth
from allianceauth.analytics.tasks import analytics_event
from allianceauth.notifications.models import Notification
from allianceauth.services.hooks import get_extension_logger
from app_utils.django import app_labels

from dens.app_settings import DENS_ADMIN_NOTIFICATIONS_ENABLED
from dens.esi import (
    TEMPERATE_PLANET_TYPE_ID,
    MissingTokenError,
    get_esi_asset_location,
    get_owner_anchored_dens_from_esi,
    get_owner_mercenarydenreinforced_notifications_from_esi,
)
from dens.models import DenOwner, MercenaryDen, MercenaryDenReinforcedNotification

logger = get_extension_logger(__name__)


class DenMissingFromDatabase(Exception):
    """Exception raised when the Den Associated with a notification can't be found in the datbaase"""


@shared_task
def update_all_den_owners():
    """Initiates an update of all enabled den owners"""
    enabled_den_owners = DenOwner.objects.enabled_owners()
    logger.info("Updating %s den owners", enabled_den_owners.count())

    task_list = []
    for owner in enabled_den_owners:
        task_list.append(update_owner_dens.si(owner.id))

    group(task_list).delay()


@shared_task
def update_owner_dens(owner_id: int):
    """Updates the mercenary dens anchored by this owner"""
    logger.info("Updating mercenary dens for owner id %s", owner_id)
    owner = DenOwner.objects.get(id=owner_id)

    try:
        dens_assets = get_owner_anchored_dens_from_esi(owner)
    except MissingTokenError as e:
        logger.error(
            "Missing token for owner %s when fetching assets. Disabling owner."
        )
        disable_owner_and_notify(owner)
        raise e

    current_ids_set = {den["item_id"] for den in dens_assets}

    stored_ids_set = MercenaryDen.objects.get_owner_dens_ids_set(owner)

    disappeared_dens_ids = stored_ids_set - current_ids_set
    disappeared_dens = MercenaryDen.objects.filter(id__in=disappeared_dens_ids)
    logger.debug("Deleting %d dens", disappeared_dens.count())
    disappeared_dens.delete()

    new_dens_ids = current_ids_set - stored_ids_set
    logger.debug("Creating %d new dens", len(new_dens_ids))
    new_dens_tasks = []
    for dens_asset in dens_assets:
        if dens_asset["item_id"] in new_dens_ids:
            create_den_tasks = create_mercenary_den.si(owner_id, dens_asset)
            new_dens_tasks.append(create_den_tasks)

    group(new_dens_tasks).delay()


@shared_task
def create_mercenary_den(owner_id: int, den_asset_dict: dict):
    """Creates a new mercenary den associated with this owner from the asset dictionary"""
    den_item_id = den_asset_dict["item_id"]
    logger.info("Creating den id %s for owner id %d", den_item_id, owner_id)
    logger.debug(den_asset_dict)
    owner = DenOwner.objects.get(id=owner_id)

    try:
        x, y, z = get_esi_asset_location(owner, den_item_id)
    except MissingTokenError as e:
        logger.error(
            "Missing ESI token for user %s when fetching asset location. Disabling owner."
        )
        disable_owner_and_notify(owner)
        raise e
    nearest_planet = nearest_celestial(
        den_asset_dict["location_id"], x, y, z, EveGroupId.PLANET
    )

    if not nearest_planet or nearest_planet.type_id != TEMPERATE_PLANET_TYPE_ID:
        raise RuntimeError(
            f"Couldn't find planet corresponding to den id {den_item_id}"
        )

    planet, _ = EvePlanet.objects.get_or_create_esi(id=nearest_planet.id)

    MercenaryDen.objects.create(owner, den_item_id, planet)


@shared_task
def update_all_owners_notifications():
    """Starts an owner update job for every owner"""
    owners = DenOwner.objects.enabled_owners()
    logger.info("Starting notifications update for %s owners", owners.count())
    jobs = []
    for owner in owners:
        jobs.append(update_owner_notifications.si(owner.id))

    group(jobs).delay()


@shared_task
def update_owner_notifications(owner_id: int):
    """Checks all notifications related to an owner and update new den reinforcement notifications"""
    logger.info("Updating notifications for owner id %s", owner_id)

    owner = get_object_or_404(DenOwner, id=owner_id)
    try:
        notifications = get_owner_mercenarydenreinforced_notifications_from_esi(owner)
    except MissingTokenError as e:
        logger.error(
            "Missing ESI token when fetching owner %s notifications. Disabling owner."
        )
        disable_owner_and_notify(owner)
        raise e

    for notification in notifications:
        if not MercenaryDenReinforcedNotification.objects.is_notification_id_known(
            notification["notification_id"]
        ):
            create_reinforce_notification.delay(notification)


@shared_task
def create_reinforce_notification(notification_json: dict):
    """Saves a reinforce notification from the ESI information"""
    logger.info("Creating a den reinforced notification from %s", notification_json)

    reinforce_notification = (
        MercenaryDenReinforcedNotification.objects.create_from_notification(
            notification_json
        )
    )

    if reinforce_notification is None:
        raise DenMissingFromDatabase(
            f"The den associated with notification {notification_json} couldn't be found"
        )

    if reinforce_notification.is_in_future():
        logger.info("Trying to add the timer to timerboards")

        application_labels = app_labels()
        if "timerboard" in application_labels:
            from dens.thirdparty.timerboard import create_timerboard_timer

            logger.debug("Timerboard detected")
            create_timerboard_timer(reinforce_notification)
            logger.info("Timer added to timerboard")

        if "structuretimers" in application_labels:
            from dens.thirdparty.structuretimers import create_structuretimer_timer

            logger.debug("Structruetimer detected")
            create_structuretimer_timer(reinforce_notification)
            logger.info("Timer added to structuretimers")

        if "aadiscordbot" in application_labels:
            from dens.thirdparty.aadiscordbot import (
                send_reinforced_notification_to_user,
            )

            logger.debug("aa-discordbot detected")
            send_reinforced_notification_to_user(reinforce_notification)
            logger.info("Sent the notification to the user through discordbot")
        else:
            logger.debug("No aa-discordbot defaulting to auth notification")
            Notification.objects.notify_user(
                user=reinforce_notification.den.owner.character_ownership.user,
                title=_("Mercenary Den reinforced"),
                message=_(
                    "Mercenary Den on %(location)s has been reinforced by %(name)s.\n"
                    "Will exit reinforcement at %(time)s."
                    % {
                        "location": reinforce_notification.den.location.name,
                        "name": reinforce_notification.reinforced_by.name,
                        "time": reinforce_notification.exit_reinforcement.isoformat(),
                    }
                ),
            )
            logger.info(
                "Sent the notification to the user through alliance auth notifications"
            )


def disable_owner_and_notify(owner: DenOwner):
    """
    Will disable an owner and try to notify a user using discord otherwise will use an alliance auth notification
    """
    logger.info("Disable den owner %s and sending disable notification to user", owner)

    owner.disable()

    if "aadiscordbot" in app_labels():
        from dens.thirdparty.aadiscordbot import send_disable_notification_to_user

        logger.debug("aa-discordbot detected")
        send_disable_notification_to_user(owner)
        logger.info("Sent the disable notification to the user through discordbot")
    else:
        logger.debug("No aa-discordbot, defaulting to auth notification")
        Notification.objects.notify_user(
            user=owner.character_ownership.user,
            title=_("Mercenary den owner disabled"),
            level=Notification.Level.WARNING,
            message=_(
                f"""Mercenary den owner {owner} has been disabled.
                "Add it back to enable services again."""
            ),
        )
        logger.info(
            "Sent the disable notification to the user through aa notifications"
        )

    if DENS_ADMIN_NOTIFICATIONS_ENABLED:
        app_utils.allianceauth.notify_admins(
            _("Mercenary Den owner disabled"),
            _(f"Mercenary Den owner {owner} has been disabled"),
        )


def send_analytics(label: str, value):
    """
    Send an analytics event
    """

    logger.info("Sending analytic %s with value %s", label, value)

    analytics_event(
        namespace="dens.analytics",
        task="send_daily_stats",
        label=label,
        value=value,
        event_type="Stats",
    )


@shared_task
def send_daily_analytics():
    """Sends analytic information to the AA GA instance"""
    logger.info("Starting the daily analytics task")

    count_owners = DenOwner.objects.count()
    count_dens = MercenaryDen.objects.count()

    send_analytics("den_owners", count_owners)
    send_analytics("mercenary_dens", count_dens)
